#![allow(dead_code)]
pub mod brain {
    tonic::include_proto!("brain");
}

use anyhow::Result;
use brain::{
    brain_daemon_service_client::BrainDaemonServiceClient, NewVmConfirmation, NewVmRequest,
    NodePubkey,
};
use log::{error, debug, warn, info};
use tokio::time::{sleep, Duration};
use tokio::{sync::mpsc::Receiver, sync::mpsc::Sender, task::JoinSet};
use tokio_stream::wrappers::ReceiverStream;
use tokio_stream::StreamExt;
use tonic::transport::Channel;

async fn listen_for_new_vm_reqs(
    mut client: BrainDaemonServiceClient<Channel>,
    tx: Sender<NewVmRequest>,
) -> Result<()> {
    let node_pubkey = "somePublicKey".to_string();
    let mut grpc_stream = client
        .get_new_vm_reqs(NodePubkey { node_pubkey })
        .await?
        .into_inner();
    while let Some(newvmreq) = grpc_stream.next().await {
        info!("Received new vm request: {newvmreq:?}");
        let _ = tx.send(newvmreq?).await;
    }
    debug!("listen_for_new_vm_reqs is about to exit");
    Ok(())
}

async fn send_confirmations(
    mut client: BrainDaemonServiceClient<Channel>,
    rx: Receiver<NewVmConfirmation>,
) -> Result<()> {
    let rx_stream = ReceiverStream::new(rx);
    client.send_vm_confirmations(rx_stream).await?;
    debug!("send_confirmations is about to exit");
    Ok(())
}

async fn connect_and_run() -> Result<()> {
    let client = BrainDaemonServiceClient::connect("http://[::1]:31337").await?;
    let mut streaming_tasks = JoinSet::new();

    let newvm_client = client.clone();
    let (tx, _newvm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(listen_for_new_vm_reqs(newvm_client, tx));

    let confirm_client = client.clone();
    let (_confirm_tx, rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(send_confirmations(confirm_client, rx));

    let task_output = streaming_tasks.join_next().await;
    warn!("One stream exited: {task_output:?}");
    Ok(())
}

async fn connection_wrapper() -> ! {
    loop {
        info!("Connecting to brain...");
        if let Err(e) = connect_and_run().await {
            error!("The connection broke: {e}");
        }
        sleep(Duration::from_secs(3)).await;
    }
}

#[tokio::main]
async fn main() {
    env_logger::builder()
        .filter_level(log::LevelFilter::Debug)
        .init();
    connection_wrapper().await;
}
